# tests/__init__.py -- Portage Unit Test functionality
# Copyright 2006-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import multiprocessing
import sys
import time
import unittest
from pathlib import Path

from unittest.runner import TextTestResult as _TextTestResult

import portage
from portage import os
from portage.util import no_color
from portage import _encodings
from portage import _unicode_decode
from portage.const import PORTAGE_PYM_PATH
from portage.output import colorize
from portage.proxy.objectproxy import ObjectProxy


# This remains constant when the real value is a mock.
EPREFIX_ORIG = portage.const.EPREFIX


class lazy_value(ObjectProxy):
    __slots__ = ("_func",)

    def __init__(self, func):
        ObjectProxy.__init__(self)
        object.__setattr__(self, "_func", func)

    def _get_target(self):
        return object.__getattribute__(self, "_func")()


@lazy_value
def cnf_path():
    if portage._not_installed:
        return os.path.join(portage.const.PORTAGE_BASE_PATH, "cnf")
    return os.path.join(
        EPREFIX_ORIG or "/", portage.const.GLOBAL_CONFIG_PATH.lstrip(os.sep)
    )


@lazy_value
def cnf_etc_path():
    if portage._not_installed:
        return str(cnf_path)
    return os.path.join(EPREFIX_ORIG or "/", "etc")


@lazy_value
def cnf_bindir():
    if portage._not_installed:
        return portage.const.PORTAGE_BIN_PATH
    return os.path.join(portage.const.EPREFIX or "/", "usr", "bin")


@lazy_value
def cnf_sbindir():
    if portage._not_installed:
        return str(cnf_bindir)
    return os.path.join(portage.const.EPREFIX or "/", "usr", "sbin")


def get_pythonpath():
    """
    Prefix current PYTHONPATH with PORTAGE_PYM_PATH, and normalize.
    """
    pythonpath = os.environ.get("PYTHONPATH")
    if pythonpath is not None and not pythonpath.strip():
        pythonpath = None
    if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
        pass
    else:
        if pythonpath is None:
            pythonpath = ""
        else:
            pythonpath = ":" + pythonpath
        pythonpath = PORTAGE_PYM_PATH + pythonpath
    return pythonpath


class TestCase(unittest.TestCase):
    """
    We need a way to mark a unit test as "ok to fail"
    This way someone can add a broken test and mark it as failed
    and then fix the code later.  This may not be a great approach
    (broken code!!??!11oneone) but it does happen at times.
    """

    def __init__(self, *pargs, **kwargs):
        unittest.TestCase.__init__(self, *pargs, **kwargs)
        self.cnf_path = cnf_path
        self.cnf_etc_path = cnf_etc_path
        self.bindir = cnf_bindir
        self.sbindir = cnf_sbindir

    def setUp(self):
        """
        Setup multiprocessing start method if needed. It needs to be
        done relatively late in order to work with the pytest-xdist
        plugin due to execnet usage.
        """
        if os.environ.get("PORTAGE_MULTIPROCESSING_START_METHOD") == "spawn":
            multiprocessing.set_start_method("spawn", force=True)

    def assertRaisesMsg(self, msg, excClass, callableObj, *args, **kwargs):
        """Fail unless an exception of class excClass is thrown
        by callableObj when invoked with arguments args and keyword
        arguments kwargs. If a different type of exception is
        thrown, it will not be caught, and the test case will be
        deemed to have suffered an error, exactly as for an
        unexpected exception.
        """
        try:
            callableObj(*args, **kwargs)
        except excClass:
            return
        else:
            if hasattr(excClass, "__name__"):
                excName = excClass.__name__
            else:
                excName = str(excClass)
            raise self.failureException(f"{excName} not raised: {msg}")

    def assertNotExists(self, path):
        """Make sure |path| does not exist"""
        path = Path(path)
        if path.exists():
            raise self.failureException(f"path exists when it should not: {path}")


test_cps = ["sys-apps/portage", "virtual/portage"]
test_versions = ["1.0", "1.0-r1", "2.3_p4", "1.0_alpha57"]
test_slots = [None, "1", "gentoo-sources-2.6.17", "spankywashere"]
test_usedeps = ["foo", "-bar", ("foo", "bar"), ("foo", "-bar"), ("foo?", "!bar?")]



# Copyright 2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class AgressiveBacktrackDowngradeTestCase(TestCase):
    def testAgressiveBacktrackDowngrade(self):
        ebuilds = {
            "www-client/firefox-69.0": {
                "EAPI": "7",
                "RDEPEND": "=media-libs/libvpx-1.7*:0=[postproc] media-video/ffmpeg",
            },
            "www-client/firefox-60.9.0": {"EAPI": "7", "RDEPEND": ""},
            "media-libs/libvpx-1.8.0": {
                "EAPI": "7",
                "SLOT": "0/6",
                "IUSE": "postproc",
            },
            "media-libs/libvpx-1.7.0": {
                "EAPI": "7",
                "SLOT": "0/5",
                "IUSE": "+postproc",
            },
            "media-libs/libvpx-1.6.0": {
                "EAPI": "7",
                "SLOT": "0/4",
                "IUSE": "postproc",
            },
            "media-video/ffmpeg-4.2": {
                "EAPI": "7",
                "RDEPEND": "media-libs/libvpx:=",
            },
        }

        installed = {
            "www-client/firefox-69.0": {
                "EAPI": "7",
                "RDEPEND": "=media-libs/libvpx-1.7*:0/5=[postproc] media-video/ffmpeg",
            },
            "media-libs/libvpx-1.7.0": {
                "EAPI": "7",
                "SLOT": "0/5",
                "IUSE": "+postproc",
                "USE": "postproc",
            },
            "media-video/ffmpeg-4.2": {
                "EAPI": "7",
                "RDEPEND": "media-libs/libvpx:0/5=",
            },
        }

        world = ["media-video/ffmpeg", "www-client/firefox"]

        test_cases = (
            # Test bug 693836, where an attempt to upgrade libvpx lead
            # to aggressive backtracking which ultimately triggered an
            # undesirable firefox downgrade like this:
            # [ebuild     U  ] media-libs/libvpx-1.8.0 [1.7.0]
            # [ebuild     UD ] www-client/firefox-60.9.0 [69.0]
            # [ebuild  rR    ] media-video/ffmpeg-4.2
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 1998-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__docformat__ = "epytext"

import errno
import itertools
import re
import socket
import subprocess
import sys
from typing import Optional

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.process:spawn",
    "portage.util:writemsg",
    "portage.util.futures:asyncio",
)
import portage.util.formatter as formatter

from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage import _unicode_decode
from portage.const import COLOR_MAP_FILE
from portage.exception import (
    CommandNotFound,
    FileNotFound,
    ParseError,
    PermissionDenied,
    PortageException,
)
from portage.localization import _

havecolor = 1
dotitles = 1

"""Maps style class to tuple of attribute names."""
_styles = {}

"""Maps attribute name to ansi code."""

esc_seq = "\x1b["

codes = {
    "normal": f"{esc_seq}0m",
    "reset": f"{esc_seq}39;49;00m",
    "bold": f"{esc_seq}01m",
    "faint": f"{esc_seq}02m",
    "standout": f"{esc_seq}03m",
    "underline": f"{esc_seq}04m",
    "blink": f"{esc_seq}05m",
    "overline": f"{esc_seq}06m",
    "reverse": f"{esc_seq}07m",
    "invisible": f"{esc_seq}08m",
    "no-attr": f"{esc_seq}22m",
    "no-standout": f"{esc_seq}23m",
    "no-underline": f"{esc_seq}24m",
    "no-blink": f"{esc_seq}25m",
    "no-overline": f"{esc_seq}26m",
    "no-reverse": f"{esc_seq}27m",
    "bg_black": f"{esc_seq}40m",
    "bg_darkred": f"{esc_seq}41m",
    "bg_darkgreen": f"{esc_seq}42m",
    "bg_brown": f"{esc_seq}43m",
    "bg_darkblue": f"{esc_seq}44m",
    "bg_purple": f"{esc_seq}45m",
    "bg_teal": f"{esc_seq}46m",
    "bg_lightgray": f"{esc_seq}47m",
    "bg_default": f"{esc_seq}49m",
    "bg_darkyellow": f"{esc_seq}43m",
}


def color(fg, bg="default", attr=["normal"]):
    myansicodechain = itertools.chain((codes[fg]), (codes[x] for x in [bg, *attr]))
    return "".join(myansicodechain)


ansi_codes = [y for x in range(30, 38) for y in (f"{x}m", f"{x};01m")]


rgb_ansi_colors = [
    "0x000000",
    "0x555555",
    "0xAA0000",
    "0xFF5555",
    "0x00AA00",
    "0x55FF55",
    "0xAA5500",
    "0xFFFF55",
    "0x0000AA",
    "0x5555FF",
    "0xAA00AA",
    "0xFF55FF",
    "0x00AAAA",
    "0x55FFFF",
    "0xAAAAAA",
    "0xFFFFFF",
]

for x in range(len(rgb_ansi_colors)):
    codes[rgb_ansi_colors[x]] = esc_seq + ansi_codes[x]

del x

codes["black"] = codes["0x000000"]
codes["darkgray"] = codes["0x555555"]

codes["red"] = codes["0xFF5555"]
codes["darkred"] = codes["0xAA0000"]

codes["green"] = codes["0x55FF55"]
codes["darkgreen"] = codes["0x00AA00"]

codes["yellow"] = codes["0xFFFF55"]
codes["brown"] = codes["0xAA5500"]

codes["blue"] = codes["0x5555FF"]
codes["darkblue"] = codes["0x0000AA"]

codes["fuchsia"] = codes["0xFF55FF"]
codes["purple"] = codes["0xAA00AA"]

codes["turquoise"] = codes["0x55FFFF"]
codes["teal"] = codes["0x00AAAA"]

codes["white"] = codes["0xFFFFFF"]
codes["lightgray"] = codes["0xAAAAAA"]

codes["darkteal"] = codes["turquoise"]
# Some terminals have darkyellow instead of brown.
codes["0xAAAA00"] = codes["brown"]
codes["darkyellow"] = codes["0xAAAA00"]


# Colors from /lib/gentoo/functions.sh
_styles["BAD"] = ("red",)
_styles["BRACKET"] = ("blue",)
_styles["ERR"] = ("red",)
_styles["GOOD"] = ("green",)
_styles["HILITE"] = ("teal",)
_styles["INFO"] = ("darkgreen",)
_styles["LOG"] = ("green",)
_styles["NORMAL"] = ("normal",)
_styles["QAWARN"] = ("brown",)
_styles["WARN"] = ("yellow",)

# Portage functions
_styles["INFORM"] = ("darkgreen",)
_styles["UNMERGE_WARN"] = ("red",)
_styles["SECURITY_WARN"] = ("red",)
_styles["MERGE_LIST_PROGRESS"] = ("yellow",)
_styles["PKG_BLOCKER"] = ("red",)
_styles["PKG_BLOCKER_SATISFIED"] = ("teal",)
_styles["PKG_MERGE"] = ("darkgreen",)
_styles["PKG_MERGE_SYSTEM"] = ("darkgreen",)
_styles["PKG_MERGE_WORLD"] = ("green",)
_styles["PKG_BINARY_MERGE"] = ("purple",)
_styles["PKG_BINARY_MERGE_SYSTEM"] = ("purple",)
_styles["PKG_BINARY_MERGE_WORLD"] = ("fuchsia",)
_styles["PKG_UNINSTALL"] = ("red",)
_styles["PKG_NOMERGE"] = ("teal",)
_styles["PKG_NOMERGE_SYSTEM"] = ("teal",)
_styles["PKG_NOMERGE_WORLD"] = ("blue",)
_styles["PROMPT_CHOICE_DEFAULT"] = ("green",)
_styles["PROMPT_CHOICE_OTHER"] = ("red",)


def _parse_color_map(config_root="/", onerror=None):
    """
    Parse /etc/portage/color.map and return a dict of error codes.

    @param onerror: an optional callback to handle any ParseError that would
            otherwise be raised
    @type onerror: callable
    @rtype: dict
    @return: a dictionary mapping color classes to color codes
    """
    global codes, _styles
    myfile = os.path.join(config_root, COLOR_MAP_FILE)
    ansi_code_pattern = re.compile("^[0-9;]*m$")
    quotes = "'\""

    def strip_quotes(token):
        if token[0] in quotes and token[0] == token[-1]:
            token = token[1:-1]
        return token

    try:
        with open(
            _unicode_encode(myfile, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["content"],
            errors="replace",
        ) as f:
            lines = f.readlines()
        for lineno, line in enumerate(lines):
            commenter_pos = line.find("#")
            line = line[:commenter_pos].strip()

            if len(line) == 0:
                continue

            split_line = line.split("=")
            if len(split_line) != 2:
                e = ParseError(
                    _("'%s', line %s: expected exactly one occurrence of '=' operator")
                    % (myfile, lineno)
                )
                raise e
                if onerror:
                    onerror(e)
                else:
                    raise e
                continue

            k = strip_quotes(split_line[0].strip())
            v = strip_quotes(split_line[1].strip())
            if not k in _styles and not k in codes:
                e = ParseError(
                    _("'%s', line %s: Unknown variable: '%s'") % (myfile, lineno, k)
                )
                if onerror:
                    onerror(e)
                else:
                    raise e
                continue
            if ansi_code_pattern.match(v):
                if k in _styles:
                    _styles[k] = (esc_seq + v,)
                elif k in codes:
                    codes[k] = esc_seq + v
            else:
                code_list = []
                for x in v.split():
                    if x in codes:
                        if k in _styles:
                            code_list.append(x)
                        elif k in codes:
                            code_list.append(codes[x])
                    else:
                        e = ParseError(
                            _("'%s', line %s: Undefined: '%s'") % (myfile, lineno, x)
                        )
                        if onerror:
                            onerror(e)
                        else:
                            raise e
                if k in _styles:
                    _styles[k] = tuple(code_list)
                elif k in codes:
                    codes[k] = "".join(code_list)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise FileNotFound(myfile)
        elif e.errno == errno.EACCES:
            raise PermissionDenied(myfile)
        raise


def nc_len(mystr):
    tmp = re.sub(esc_seq + "^m]+m", "", mystr)
    return len(tmp)


_legal_terms_re = re.compile(
    r"^(xterm|xterm-color|Eterm|aterm|rxvt|screen|kterm|rxvt-unicode|gnome|interix|tmux|st-256color|alacritty|konsole|foot)"
)
_disable_xtermTitle = None
_max_xtermTitle_len = 253


def xtermTitle(mystr, raw=False):
    global _disable_xtermTitle
    if _disable_xtermTitle is None:
        _disable_xtermTitle = not (
            sys.__stderr__.isatty()
            and "TERM" in os.environ
            and _legal_terms_re.match(os.environ["TERM"]) is not None
        )

    if dotitles and not _disable_xtermTitle:
        if not raw:
            hostname = socket.gethostname()
            mystr = f"{hostname}: {mystr}"
        # If the title string is too big then the terminal can
        # misbehave. Therefore, truncate it if it's too big.
        if len(mystr) > _max_xtermTitle_len:
            mystr = mystr[:_max_xtermTitle_len]
        if not raw:
            mystr = f"\x1b]0;{mystr}\x07"

        # avoid potential UnicodeEncodeError
        mystr = _unicode_encode(
            mystr, encoding=_encodings["stdio"], errors="backslashreplace"
        )
        f = sys.stderr.buffer
        f.write(mystr)
        f.flush()


default_xterm_title = None


def xtermTitleReset():
    global default_xterm_title
    if default_xterm_title is None:
        prompt_command = os.environ.get("PROMPT_COMMAND")
        if prompt_command == "":
            default_xterm_title = ""
        elif prompt_command is not None:
            if (
                dotitles
                and "TERM" in os.environ
                and _legal_terms_re.match(os.environ["TERM"]) is not None
                and sys.__stderr__.isatty()
            ):
                from portage.process import find_binary, spawn

                shell = os.environ.get("SHELL")
                if not shell or not os.access(shell, os.EX_OK):
                    shell = find_binary("sh")
                if shell:
                    spawn(
                        [shell, "-c", prompt_command],
                        env=os.environ,
                        fd_pipes={
                            0: portage._get_stdin().fileno(),
                            1: sys.__stderr__.fileno(),
                            2: sys.__stderr__.fileno(),
                        },
                    )
                else:
                    os.system(prompt_command)
            return
        else:
            pwd = os.environ.get("PWD", "")
            home = os.environ.get("HOME", "")
            if home != "" and pwd.startswith(home):
                pwd = "~" + pwd[len(home) :]
            default_xterm_title = "\x1b]0;{}@{}:{}\x07".format(
                os.environ.get("LOGNAME", ""),
                os.environ.get("HOSTNAME", "").split(".", 1)[0],
                pwd,
            )
    xtermTitle(default_xterm_title, raw=True)


def notitles():
    "turn off title setting"
    dotitles = 0


def nocolor():
    "turn off colorization"
    global havecolor
    havecolor = 0


def resetColor():
    return codes["reset"]


def style_to_ansi_code(style):
    """
    @param style: A style name
    @type style: String
    @rtype: String
    @return: A string containing one or more ansi escape codes that are
            used to render the given style.
    """
    ret = ""
    for attr_name in _styles[style]:
        # allow stuff that has found it's way through ansi_code_pattern
        ret += codes.get(attr_name, attr_name)
    return ret


def colormap():
    mycolors = []
    for c in (
        "BAD",
        "BRACKET",
        "ERR",
        "GOOD",
        "HILITE",
        "INFO",
        "LOG",
        "NORMAL",
        "QAWARN",
        "WARN",
    ):
        mycolors.append(f"PORTAGE_COLOR_{c}=$'{style_to_ansi_code(c)}'")
    return "\n".join(mycolors)


def colorize(color_key, text):
    global havecolor
    if havecolor:
        if color_key in codes:
            return codes[color_key] + text + codes["reset"]
        if color_key in _styles:
            return style_to_ansi_code(color_key) + text + codes["reset"]
    return text


compat_functions_colors = [
    "bold",
    "white",
    "teal",
    "turquoise",
    "darkteal",
    "fuchsia",
    "purple",
    "blue",
    "darkblue",
    "green",
    "darkgreen",
    "yellow",
    "brown",
    "darkyellow",
    "red",
    "darkred",
]


class create_color_func:
    __slots__ = ("_color_key",)

    def __init__(self, color_key):
        self._color_key = color_key

    def __call__(self, text):
        return colorize(self._color_key, text)


for c in compat_functions_colors:
    globals()[c] = create_color_func(c)


class ConsoleStyleFile:
    """
    A file-like object that behaves something like
    the colorize() function. Style identifiers
    passed in via the new_styles() method will be used to
    apply console codes to output.
    """

    def __init__(self, f):
        self._file = f
        self._styles = None
        self.write_listener = None

    def new_styles(self, styles):
        self._styles = styles

    def write(self, s):
        # In python-2.6, DumbWriter.send_line_break() can write
        # non-unicode '\n' which fails with TypeError if self._file
        # is a text stream such as io.StringIO. Therefore, make sure
        # input is converted to unicode when necessary.
        s = _unicode_decode(s)
        global havecolor
        if havecolor and self._styles:
            styled_s = []
            for style in self._styles:
                styled_s.append(style_to_ansi_code(style))
            styled_s.append(s)
            styled_s.append(codes["reset"])
            self._write(self._file, "".join(styled_s))
        else:
            self._write(self._file, s)
        if self.write_listener:
            self._write(self.write_listener, s)

    def _write(self, f, s):
        # avoid potential UnicodeEncodeError
        if f in (sys.stdout, sys.stderr):
            s = _unicode_encode(
                s, encoding=_encodings["stdio"], errors="backslashreplace"
            )
            f = f.buffer
        f.write(s)

    def writelines(self, lines):
        for s in lines:
            self.write(s)

    def flush(self):
        self._file.flush()

    def close(self):
        self._file.close()


class StyleWriter(formatter.DumbWriter):
    """
    This is just a DumbWriter with a hook in the new_styles() method
    that passes a styles tuple as a single argument to a callable
    style_listener attribute.
    """

    def __init__(self, **kwargs):
        formatter.DumbWriter.__init__(self, **kwargs)
        self.style_listener = None

    def new_styles(self, styles):
        formatter.DumbWriter.new_styles(self, styles)
        if self.style_listener:
            self.style_listener(styles)


def get_term_size(fd=None):
    """
    Get the number of lines and columns of the tty that is connected to
    fd.  Returns a tuple of (lines, columns) or (0, 0) if an error
    occurs. The curses module is used if available, otherwise the output of
    `stty size` is parsed. The lines and columns values are guaranteed to be
    greater than or equal to zero, since a negative COLUMNS variable is
    known to prevent some commands from working (see bug #394091).
    """
    if fd is None:
        fd = sys.stdout
    if not hasattr(fd, "isatty") or not fd.isatty():
        return (0, 0)

    # Do not use curses.tigetnum("lines") or curses.tigetnum("cols") because it
    # returns stale values after terminal resize. Do not use curses.initscr().getmaxyx()
    # since that has unwanted side-effects, requiring a call to `stty sane` to restore a
    # sane state.
    if not sys.stdin.isatty():
        # Use curses if stdin is not a tty since stty fails in that case.
        try:
            import curses

            try:
                curses.setupterm(term=os.environ.get("TERM", "unknown"), fd=fd.fileno())
                return curses.tigetnum("lines"), curses.tigetnum("cols")
            except curses.error:
                pass
        except ImportError:
            pass

    try:
        proc = subprocess.Popen(["stty", "size"], stdout=subprocess.PIPE, stderr=fd)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        # stty command not found
        return (0, 0)

    out = _unicode_decode(proc.communicate()[0])
    if proc.wait() == os.EX_OK:
        out = out.split()
        if len(out) == 2:
            try:
                val = (int(out[0]), int(out[1]))
            except ValueError:
                pass
            else:
                if val[0] >= 0 and val[1] >= 0:
                    return val
    return (0, 0)


def set_term_size(lines: int, columns: int, fd: int) -> Optional[asyncio.Future]:
    """
    Set the number of lines and columns for the tty that is connected to fd.
    For portability, this simply calls `stty rows $lines columns $columns`.

    If spawn succeeds and the event loop is running then an instance of
    asyncio.Future is returned and the caller should wait for it in order
    to prevent possible error messages like this:

    [ERROR] Task was destroyed but it is pending!
    """

    cmd = ["stty", "rows", str(lines), "columns", str(columns)]
    try:
        proc = spawn(cmd, env=os.environ, fd_pipes={0: fd}, returnproc=True)
    except CommandNotFound:
        writemsg(_("portage: stty: command not found\n"), noiselevel=-1)
    else:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            return asyncio.ensure_future(proc.wait(), loop=loop)
        else:
            loop.run_until_complete(proc.wait())


class EOutput:
    """
    Performs fancy terminal formatting for status and informational messages.

    The provided methods produce identical terminal output to the eponymous
    functions in the shell script C{/sbin/functions.sh} and also accept
    identical parameters.

    This is not currently a drop-in replacement however, as the output-related
    functions in C{/sbin/functions.sh} are oriented for use mainly by system
    init scripts and ebuilds and their output can be customized via certain
    C{RC_*} environment variables (see C{/etc/conf.d/rc}). B{EOutput} is not
    customizable in this manner since it's intended for more general uses.
    Likewise, no logging is provided.

    @ivar quiet: Specifies if output should be silenced.
    @type quiet: BooleanType
    @ivar term_columns: Width of terminal in characters. Defaults to the value
            specified by the shell's C{COLUMNS} variable, else to the queried tty
            size, else to C{80}.
    @type term_columns: IntType
    """

    def __init__(self, quiet=False):
        self.__last_e_cmd = ""
        self.__last_e_len = 0
        self.quiet = quiet
        lines, columns = get_term_size()
        if columns <= 0:
            columns = 80
        self.term_columns = columns
        sys.stdout.flush()
        sys.stderr.flush()

    def _write(self, f, s):
        # avoid potential UnicodeEncodeError
        writemsg(s, noiselevel=-1, fd=f)

    def __eend(self, caller, errno, msg):
        if errno == 0:
            status_brackets = (
                colorize("BRACKET", "[ ")
                + colorize("GOOD", "ok")
                + colorize("BRACKET", " ]")
            )
        else:
            status_brackets = (
                colorize("BRACKET", "[ ")
                + colorize("BAD", "!!")
                + colorize("BRACKET", " ]")
            )
            if msg:
                if caller == "eend":
                    self.eerror(msg[0])
                elif caller == "ewend":
                    self.ewarn(msg[0])
        if self.__last_e_cmd != "ebegin":
            self.__last_e_len = 0
        if not self.quiet:
            out = sys.stdout
            self._write(
                out,
                "%*s%s\n"
                % ((self.term_columns - self.__last_e_len - 7), "", status_brackets),
            )

    def ebegin(self, msg):
        """
        Shows a message indicating the start of a process.

        @param msg: A very brief (shorter than one line) description of the
                starting process.
        @type msg: StringType
        """
        msg += " ..."
        if not self.quiet:
            self.einfon(msg)
        self.__last_e_len = len(msg) + 3
        self.__last_e_cmd = "ebegin"

    def eend(self, errno, *msg):
        """
        Indicates the completion of a process, optionally displaying a message
        via L{eerror} if the process's exit status isn't C{0}.

        @param errno: A standard UNIX C{errno} code returned by processes upon
                exit.
        @type errno: IntType
        @param msg: I{(optional)} An error message, typically a standard UNIX
                error string corresponding to C{errno}.
        @type msg: StringType
        """
        if not self.quiet:
            self.__eend("eend", errno, msg)
        self.__last_e_cmd = "eend"

    def eerror(self, msg):
        """
        Shows an error message.

        @param msg: A very brief (shorter than one line) error message.
        @type msg: StringType
        """
        out = sys.stderr
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("ERR", " * ") + msg + "\n")
        self.__last_e_cmd = "eerror"

    def einfo(self, msg):
        """
        Shows an informative message terminated with a newline.

        @param msg: A very brief (shorter than one line) informative message.
        @type msg: StringType
        """
        out = sys.stdout
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("INFO", " * ") + msg + "\n")
        self.__last_e_cmd = "einfo"

    def einfon(self, msg):
        """
        Shows an informative message terminated without a newline.

        @param msg: A very brief (shorter than one line) informative message.
        @type msg: StringType
        """
        out = sys.stdout
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("INFO", " * ") + msg)
        self.__last_e_cmd = "einfon"

    def eqawarn(self, msg):
        """
        Shows a QA warning.

        @param msg: A very brief (shorter than one line) warning message.
        @type msg: StringType
        """
        out = sys.stderr
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("QAWARN", " * ") + msg + "\n")
        self.__last_e_cmd = "ewarn"

    def elog(self, msg):
        """
        Shows a logged informative message terminated with a newline.

        @param msg: A very brief (shorter than one line) informative message.
        @type msg: StringType
        """
        out = sys.stdout
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("LOG", " * ") + msg + "\n")
        self.__last_e_cmd = "elog"

    def ewarn(self, msg):
        """
        Shows a warning message.

        @param msg: A very brief (shorter than one line) warning message.
        @type msg: StringType
        """
        out = sys.stderr
        if not self.quiet:
            if self.__last_e_cmd == "ebegin":
                self._write(out, "\n")
            self._write(out, colorize("WARN", " * ") + msg + "\n")
        self.__last_e_cmd = "ewarn"

    def ewend(self, errno, *msg):
        """
        Indicates the completion of a process, optionally displaying a message
        via L{ewarn} if the process's exit status isn't C{0}.

        @param errno: A standard UNIX C{errno} code returned by processes upon
                exit.
        @type errno: IntType
        @param msg: I{(optional)} A warning message, typically a standard UNIX
                error string corresponding to C{errno}.
        @type msg: StringType
        """
        if not self.quiet:
            self.__eend("ewend", errno, msg)
        self.__last_e_cmd = "ewend"


class ProgressBar:
    """The interface is copied from the ProgressBar class from the EasyDialogs
    module (which is Mac only)."""

    def __init__(self, title=None, maxval=0, label=None, max_desc_length=25):
        self._title = title or ""
        self._maxval = maxval
        self._label = label or ""
        self._curval = 0
        self._desc = ""
        self._desc_max_length = max_desc_length
        self._set_desc()

    @property
    def curval(self):
        """
        The current value (of type integer or long integer) of the progress
        bar. The normal access methods coerce curval between 0 and maxval. This
        attribute should not be altered directly.
        """
        return self._curval

    @property
    def maxval(self):
        """
        The maximum value (of type integer or long integer) of the progress
        bar; the progress bar (thermometer style) is full when curval equals
        maxval. If maxval is 0, the bar will be indeterminate (barber-pole).
        This attribute should not be altered directly.
        """
        return self._maxval

    def title(self, newstr):
        """Sets the text in the title bar of the progress dialog to newstr."""
        self._title = newstr
        self._set_desc()

    def label(self, newstr):
        """Sets the text in the progress box of the progress dialog to newstr."""
        self._label = newstr
        self._set_desc()

    def _set_desc(self):
        self._desc = "{}{}".format(
            f"{self._title}: " if self._title else "",
            f"{self._label}" if self._label else "",
        )
        if len(self._desc) > self._desc_max_length:  # truncate if too long
            self._desc = f"{self._desc[:self._desc_max_length - 3]}..."
        if len(self._desc):
            self._desc = self._desc.ljust(self._desc_max_length)

    def set(self, value, maxval=None):
        """
        Sets the progress bar's curval to value, and also maxval to max if the
        latter is provided. value is first coerced between 0 and maxval. The
        thermometer bar is updated to reflect the changes, including a change
        from indeterminate to determinate or vice versa.
        """
        if maxval is not None:
            self._maxval = maxval
        if value < 0:
            value = 0
        elif value > self._maxval:
            value = self._maxval
        self._curval = value

    def inc(self, n=1):
        """Increments the progress bar's curval by n, or by 1 if n is not
        provided. (Note that n may be negative, in which case the effect is a
        decrement.) The progress bar is updated to reflect the change. If the
        bar is indeterminate, this causes one ``spin'' of the barber pole. The
        resulting curval is coerced between 0 and maxval if incrementing causes
        it to fall outside this range.
        """
        self.set(self._curval + n)


class TermProgressBar(ProgressBar):
    """A tty progress bar similar to wget's."""

    def __init__(self, fd=sys.stdout, **kwargs):
        ProgressBar.__init__(self, **kwargs)
        lines, self.term_columns = get_term_size(fd)
        self.file = fd
        self._min_columns = 11
        self._max_columns = 80
        # for indeterminate mode, ranges from 0.0 to 1.0
        self._position = 0.0

    def set(self, value, maxval=None):
        ProgressBar.set(self, value, maxval=maxval)
        self._display_image(self._create_image())

    def _display_image(self, image):
        self.file.write("\r")
        self.file.write(image)
        self.file.flush()

    def _create_image(self):
        cols = self.term_columns
        if cols > self._max_columns:
            cols = self._max_columns
        min_columns = self._min_columns
        curval = self._curval
        maxval = self._maxval
        position = self._position
        percentage_str_width = 5
        square_brackets_width = 2
        if cols < percentage_str_width:
            return ""
        bar_space = cols - percentage_str_width - square_brackets_width - 1
        if self._desc:
            bar_space -= self._desc_max_length

        if maxval == 0:
            max_bar_width = bar_space - 3
            _percent = "".ljust(percentage_str_width)
            if cols < min_columns:
                return ""
            if position <= 0.5:
                offset = 2 * position
            else:
                offset = 2 * (1 - position)
            delta = 0.5 / max_bar_width
            position += delta
            if position >= 1.0:
                position = 0.0
            # make sure it touches the ends
            if 1.0 - position < delta:
                position = 1.0
            if position < 0.5 and 0.5 - position < delta:
                position = 0.5
            self._position = position
            bar_width = int(offset * max_bar_width)
            image = "{}{}{}".format(
                self._desc,
                _percent,
                "["
                + (bar_width * " ")
                + "<=>"
                + ((max_bar_width - bar_width) * " ")
                + "]",
            )
            return image

        percentage = 100 * curval // maxval
        max_bar_width = bar_space - 1
        _percent = ("%d%% " % percentage).rjust(percentage_str_width)
        image = f"{self._desc}{_percent}"

        if cols < min_columns:
            return image
        offset = curval / maxval
        bar_width = int(offset * max_bar_width)
        image = (
            image
            + "["
            + (bar_width * "=")
            + ">"
            + ((max_bar_width - bar_width) * " ")
            + "]"
        )
        return image


_color_map_loaded = False


def _init(config_root="/"):
    """
    Load color.map from the given config_root. This is called automatically
    on first access of the codes or _styles attributes (unless it has already
    been called for some other reason).
    """

    global _color_map_loaded, codes, _styles
    if _color_map_loaded:
        return

    _color_map_loaded = True
    codes = object.__getattribute__(codes, "_attr")
    _styles = object.__getattribute__(_styles, "_attr")

    for k, v in codes.items():
        codes[k] = _unicode_decode(v)

    for k, v in _styles.items():
        _styles[k] = _unicode_decode(v)

    try:
        _parse_color_map(
            config_root=config_root,
            onerror=lambda e: writemsg(f"{str(e)}\n", noiselevel=-1),
        )
    except FileNotFound:
        pass
    except PermissionDenied as e:
        writemsg(_("Permission denied: '%s'\n") % str(e), noiselevel=-1)
        del e
    except PortageException as e:
        writemsg(f"{str(e)}\n", noiselevel=-1)
        del e


class _LazyInitColorMap(portage.proxy.objectproxy.ObjectProxy):
    __slots__ = ("_attr",)

    def __init__(self, attr):
        portage.proxy.objectproxy.ObjectProxy.__init__(self)
        object.__setattr__(self, "_attr", attr)

    def _get_target(self):
        _init()
        return object.__getattribute__(self, "_attr")


codes = _LazyInitColorMap(codes)
_styles = _LazyInitColorMap(_styles)



# Copyright 2013-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SimpleDepcleanTestCase(TestCase):
    def testSimpleDepclean(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/B:=",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/A",
            },
            "dev-libs/C-1": {},
        }

        installed = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/B:0/0=",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/A",
            },
            "dev-libs/C-1": {},
        }

        world = ("dev-libs/C",)

        test_cases = (
            # Remove dev-libs/A-1 first because of dev-libs/B:0/0= (built
            # slot-operator dep).
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                ordered=True,
                cleanlist=["dev-libs/A-1", "dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testIDEPENDDepclean(self):
        """
        Test for bug 916135, where a direct circular dependency caused
        the unmerge order to fail to account for IDEPEND.
        """

        ebuilds = {
            "dev-util/A-1": {},
            "dev-libs/B-1": {
                "EAPI": "8",
                "IDEPEND": "dev-util/A",
                "RDEPEND": "dev-libs/B:=",
            },
            "dev-libs/C-1": {},
        }

        installed = {
            "dev-util/A-1": {},
            "dev-libs/B-1": {
                "EAPI": "8",
                "IDEPEND": "dev-util/A",
                "RDEPEND": "dev-libs/B:0/0=",
            },
            "dev-libs/C-1": {},
        }

        world = ("dev-libs/C",)

        test_cases = (
            # Remove dev-libs/B first because it IDEPENDs on dev-util/A
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                ordered=True,
                cleanlist=[
                    "dev-libs/B-1",
                    "dev-util/A-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testCircularDepclean(self):
        """
        Test for bug 916135, where an indirect circular dependency caused
        the unmerge order to fail to account for IDEPEND.
        """

        ebuilds = {
            "dev-util/A-1": {},
            "dev-libs/B-1": {
                "EAPI": "8",
                "SLOT": "1",
                "IDEPEND": "dev-util/A",
                "RDEPEND": "dev-libs/B:=",
            },
            "dev-libs/B-2": {
                "EAPI": "8",
                "SLOT": "2",
                "IDEPEND": "dev-util/A",
                "RDEPEND": "dev-libs/B:=",
            },
            "dev-libs/C-1": {},
        }

        installed = {
            "dev-util/A-1": {},
            "dev-libs/B-1": {
                "EAPI": "8",
                "SLOT": "1",
                "IDEPEND": "dev-util/A",
                "RDEPEND": "dev-libs/B:2/2=",
            },
            "dev-libs/B-2": {
                "EAPI": "8",
                "SLOT": "2",
                "IDEPEND": "dev-util/A",
                "RDEPEND": "dev-libs/B:1/1=",
            },
            "dev-libs/C-1": {},
        }

        world = ("dev-libs/C",)

        test_cases = (
            # Remove dev-libs/B first because it IDEPENDs on dev-util/A
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                ordered=True,
                cleanlist=["dev-libs/B-2", "dev-libs/B-1", "dev-util/A-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2014-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import shlex

import portage

portage._internal_caller = True
portage._sync_mode = True
from portage.output import bold, red, create_color_func
from portage._global_updates import _global_updates
from portage.sync.controller import SyncManager
from portage.util.digraph import digraph
from portage.util.futures import asyncio
from portage.util._async.AsyncScheduler import AsyncScheduler

import _emerge
from _emerge.emergelog import emergelog


portage.proxy.lazyimport.lazyimport(
    globals(),
    "_emerge.actions:adjust_configs,load_emerge_config",
    "_emerge.chk_updated_cfg_files:chk_updated_cfg_files",
    "_emerge.main:parse_opts",
    "_emerge.post_emerge:display_news_notification",
)

warn = create_color_func("WARN")


class SyncRepos:
    short_desc = "Check repos.conf settings and/or sync repositories"

    @staticmethod
    def name():
        return "sync"

    def can_progressbar(self, func):
        return False

    def __init__(self, emerge_config=None, emerge_logging=False):
        """Class init function

        @param emerge_config: optional an emerge_config instance to use
        @param emerge_logging: boolean, defaults to False
        """
        if emerge_config is None:
            # need a basic options instance
            actions, opts, _files = parse_opts([], silent=True)
            emerge_config = load_emerge_config(action="sync", args=_files, opts=opts)

            # Parse EMERGE_DEFAULT_OPTS, for settings like
            # --package-moves=n.
            cmdline = shlex.split(
                emerge_config.target_config.settings.get("EMERGE_DEFAULT_OPTS", "")
            )
            emerge_config.opts = parse_opts(cmdline, silent=True)[1]

            if hasattr(portage, "settings"):
                # cleanly destroy global objects
                portage._reset_legacy_globals()
                # update redundant global variables, for consistency
                # and in order to conserve memory
                portage.settings = emerge_config.target_config.settings
                portage.db = emerge_config.trees
                portage.root = portage.db._target_eroot

        self.emerge_config = emerge_config
        if emerge_logging:
            _emerge.emergelog._disable = False
        self.xterm_titles = (
            "notitles" not in self.emerge_config.target_config.settings.features
        )
        emergelog(self.xterm_titles, " === sync")

    def auto_sync(self, **kwargs):
        """Sync auto-sync enabled repos"""
        options = kwargs.get("options", None)
        if options:
            return_messages = options.get("return-messages", False)
        else:
            return_messages = False
        success, repos, msgs = self._get_repos(auto_sync_only=True)
        if not success:
            if return_messages:
                return (False, msgs)
            return (False, None)
        return self._sync(repos, return_messages, emaint_opts=options)

    def all_repos(self, **kwargs):
        """Sync all repos defined in repos.conf"""
        options = kwargs.get("options", None)
        if options:
            return_messages = options.get("return-messages", False)
        else:
            return_messages = False
        success, repos, msgs = self._get_repos(auto_sync_only=False)
        if not success:
            if return_messages:
                return (False, msgs)
            return (False, None)
        return self._sync(repos, return_messages, emaint_opts=options)

    def repo(self, **kwargs):
        """Sync the specified repo"""
        options = kwargs.get("options", None)
        if options:
            repo_names = options.get("repo", "")
            return_messages = options.get("return-messages", False)
        else:
            return_messages = False
        if isinstance(repo_names, str):
            repo_names = repo_names.split()
        success, repos, msgs = self._get_repos(
            auto_sync_only=False, match_repos=repo_names
        )
        if not success:
            if return_messages:
                return (False, msgs)
            return (False, None)
        return self._sync(repos, return_messages, emaint_opts=options)

    @staticmethod
    def _match_repos(repos, available):
        """Internal search, matches up the repo name or alias in repos.

        @param repos: list of repo names or aliases to match
        @param available: list of repo objects to search
        @return: list of repo objects that match
        """
        selected = []
        for repo in available:
            if repo.name in repos:
                selected.append(repo)
            elif repo.aliases is not None and any(
                alias in repos for alias in repo.aliases
            ):
                selected.append(repo)
        return selected

    def _get_repos(self, auto_sync_only=True, match_repos=None):
        msgs = []
        repos = self.emerge_config.target_config.settings.repositories
        if match_repos is not None:
            # Discard duplicate repository names or aliases.
            match_repos = set(match_repos)
            repos = self._match_repos(match_repos, repos)
            if len(repos) < len(match_repos):
                # Build a set of all the matched repos' names and aliases so we
                # can do a set difference for names that are missing.
                repo_names = set()
                for repo in repos:
                    repo_names.add(repo.name)
                    if repo.aliases is not None:
                        repo_names.update(repo.aliases)
                missing = match_repos - repo_names
                if missing:
                    msgs.append(
                        red(" * ")
                        + "The specified repo(s) were not found: %s"
                        % (" ".join(repo_name for repo_name in missing))
                        + "\n   ...returning"
                    )
                    return (False, repos, msgs)

        if auto_sync_only:
            repos = self._filter_auto(repos)

        sync_disabled = [repo for repo in repos if repo.sync_type is None]
        if sync_disabled:
            repos = [repo for repo in repos if repo.sync_type is not None]
            if match_repos is not None:
                msgs.append(
                    red(" * ")
                    + "The specified repo(s) have sync disabled: %s"
                    % " ".join(repo.name for repo in sync_disabled)
                    + "\n   ...returning"
                )
                return (False, repos, msgs)

        missing_sync_uri = [repo for repo in repos if repo.sync_uri is None]
        if missing_sync_uri:
            repos = [repo for repo in repos if repo.sync_uri is not None]
            msgs.append(
                red(" * ")
                + "The specified repo(s) are missing sync-uri: %s"
                % " ".join(repo.name for repo in missing_sync_uri)
                + "\n   ...returning"
            )
            return (False, repos, msgs)

        return (True, repos, msgs)

    def _filter_auto(self, repos):
        selected = []
        for repo in repos:
            if repo.auto_sync in ["yes", "true"]:
                selected.append(repo)
        return selected

    def _sync(self, selected_repos, return_messages, emaint_opts=None):
        msgs = []
        if not selected_repos:
            if return_messages:
                msgs.append("Nothing to sync... returning")
                return (True, msgs)
            return (True, None)

        if emaint_opts is not None:
            for k, v in emaint_opts.items():
                if v is not None:
                    k = "--" + k.replace("_", "-")
                    self.emerge_config.opts[k] = v

        # Portage needs to ensure a sane umask for the files it creates.
        os.umask(0o22)

        sync_manager = SyncManager(self.emerge_config.target_config.settings, emergelog)

        max_jobs = (
            self.emerge_config.opts.get("--jobs", 1)
            if "parallel-fetch" in self.emerge_config.target_config.settings.features
            else 1
        )
        sync_scheduler = SyncScheduler(
            emerge_config=self.emerge_config,
            selected_repos=selected_repos,
            sync_manager=sync_manager,
            max_jobs=max_jobs,
            event_loop=asyncio._safe_loop(),
        )

        sync_scheduler.start()
        sync_scheduler.wait()
        retvals = sync_scheduler.retvals
        msgs.extend(sync_scheduler.msgs)
        returncode = True

        if retvals:
            msgs.extend(self.rmessage(retvals, "sync"))
            for repo, retval in retvals:
                if retval != os.EX_OK:
                    returncode = False
                    break
        else:
            msgs.extend(self.rmessage([("None", os.EX_OK)], "sync"))

        # run the post_sync_hook one last time for
        # run only at sync completion hooks
        if sync_scheduler.global_hooks_enabled:
            rcode = sync_manager.perform_post_sync_hook("")
            if rcode:
                msgs.extend(self.rmessage([("None", rcode)], "post-sync"))
                if rcode != os.EX_OK:
                    returncode = False

        # Reload the whole config.
        portage._sync_mode = False
        self._reload_config()
        self._do_pkg_moves()
        msgs.extend(self._check_updates())
        display_news_notification(
            self.emerge_config.target_config, self.emerge_config.opts
        )

        if return_messages:
            return (returncode, msgs)
        return (returncode, None)

    def _do_pkg_moves(self):
        configs = [self.emerge_config.target_config]
        if (
            self.emerge_config.target_config.root
            != self.emerge_config.running_config.root
        ):
            configs.append(self.emerge_config.running_config)
        for root_config in configs:
            if self.emerge_config.opts.get(
                "--package-moves"
            ) != "n" and _global_updates(
                root_config.root,
                self.emerge_config.trees,
                root_config.mtimedb["updates"],
                quiet=("--quiet" in self.emerge_config.opts),
            ):
                root_config.mtimedb.commit()
                # Reload the whole config.
                self._reload_config()

    def _check_updates(self):
        mybestpv = self.emerge_config.target_config.trees["porttree"].dbapi.xmatch(
            "bestmatch-visible", portage.const.PORTAGE_PACKAGE_ATOM
        )
        mypvs = portage.best(
            self.emerge_config.target_config.trees["vartree"].dbapi.match(
                portage.const.PORTAGE_PACKAGE_ATOM
            )
        )
        try:
            old_use = (
                self.emerge_config.target_config.trees["vartree"]
                .dbapi.aux_get(mypvs, ["USE"])[0]
                .split()
            )
        except KeyError:
            old_use = ()

        chk_updated_cfg_files(
            self.emerge_config.target_config.root,
            self.emerge_config.target_config.settings.get("CONFIG_PROTECT", "").split(),
        )

        msgs = []
        if (
            not (mybestpv and mypvs)
            or mybestpv == mypvs
            or "--quiet" in self.emerge_config.opts
        ):
            return msgs

        # Suggest to update to the latest available version of portage.
        # Since changes to PYTHON_TARGETS cause complications, this message
        # is suppressed if the new version has different PYTHON_TARGETS enabled
        # than previous version.
        portdb = self.emerge_config.target_config.trees["porttree"].dbapi
        portdb.doebuild_settings.setcpv(mybestpv, mydb=portdb)
        usemask = portdb.doebuild_settings.usemask
        useforce = portdb.doebuild_settings.useforce
        new_use = (
            frozenset(portdb.doebuild_settings["PORTAGE_USE"].split()) | useforce
        ) - usemask
        new_python_targets = frozenset(
            x for x in new_use if x.startswith("python_targets_")
        )
        old_python_targets = frozenset(
            x for x in old_use if x.startswith("python_targets_")
        )

        if new_python_targets == old_python_targets:
            msgs.append("")
            msgs.append(
                warn(" * ")
                + bold("An update to portage is available.")
                + " It is _highly_ recommended"
            )
            msgs.append(
                warn(" * ")
                + "that you update portage now, before any other packages are updated."
            )
            msgs.append("")
            msgs.append(
                warn(" * ")
                + "To update portage, run 'emerge --oneshot sys-apps/portage' now."
            )
            msgs.append("")
        return msgs

    def _reload_config(self):
        """Reload the whole config from scratch."""
        load_emerge_config(emerge_config=self.emerge_config)
        adjust_configs(self.emerge_config.opts, self.emerge_config.trees)

    def rmessage(self, rvals, action):
        """Creates emaint style messages to return to the task handler"""
        messages = []
        for rval in rvals:
            messages.append(
                f"Action: {action} for repo: {rval[0]}, returned code = {rval[1]}"
            )
        return messages


class SyncScheduler(AsyncScheduler):
    """
    Sync repos in parallel, but don't sync a given repo until all
    of its masters have synced.
    """

    def __init__(self, **kwargs):
        """
        @param emerge_config: an emerge_config instance
        @param selected_repos: list of RepoConfig instances
        @param sync_manager: a SyncManger instance
        """
        self._emerge_config = kwargs.pop("emerge_config")
        self._selected_repos = kwargs.pop("selected_repos")
        self._sync_manager = kwargs.pop("sync_manager")
        AsyncScheduler.__init__(self, **kwargs)
        self._init_graph()
        self.retvals = []
        self.msgs = []

    def _init_graph(self):
        """
        Graph relationships between repos and their masters.
        """
        self._sync_graph = digraph()
        self._leaf_nodes = []
        self._repo_map = {}
        self._running_repos = set()
        selected_repo_names = frozenset(repo.name for repo in self._selected_repos)
        for repo in self._selected_repos:
            self._repo_map[repo.name] = repo
            self._sync_graph.add(repo.name, None)
            for master in repo.masters:
                if master.name in selected_repo_names:
                    self._repo_map[master.name] = master
                    self._sync_graph.add(master.name, repo.name)
        self._complete_graph = self._sync_graph.copy()
        self._hooks_repos = set()
        self._update_leaf_nodes()

    def _task_exit(self, task):
        """
        Remove the task from the graph, in order to expose
        more leaf nodes.
        """
        self._running_tasks.discard(task)
        # Set hooks_enabled = True by default, in order to ensure
        # that hooks will be called in a backward-compatible manner
        # even if all sync tasks have failed.
        hooks_enabled = True
        returncode = task.returncode
        if task.returncode == os.EX_OK:
            returncode, message, updatecache_flg, hooks_enabled = task.result
            if message:
                self.msgs.append(message)
        repo = task.kwargs["repo"].name
        self._running_repos.remove(repo)
        self.retvals.append((repo, returncode))
        self._sync_graph.remove(repo)
        self._update_leaf_nodes()
        if hooks_enabled:
            self._hooks_repos.add(repo)
        super()._task_exit(self)

    def _master_hooks(self, repo_name):
        """
        @param repo_name: a repo name
        @type repo_name: str
        @return: True if hooks would have been executed for any master
                repositories of the given repo, False otherwise
        @rtype: bool
        """
        traversed_nodes = set()
        node_stack = [repo_name]
        while node_stack:
            node = node_stack.pop()
            if node in self._hooks_repos:
                return True
            if node not in traversed_nodes:
                traversed_nodes.add(node)
                node_stack.extend(self._complete_graph.child_nodes(node))
        return False

    @property
    def global_hooks_enabled(self):
        """
        @return: True if repo.postsync.d hooks would have been executed
                for any repositories.
        @rtype: bool
        """
        return bool(self._hooks_repos)

    def _update_leaf_nodes(self):
        """
        Populate self._leaf_nodes with current leaves from
        self._sync_graph. If a circular master relationship
        is discovered, choose a random node to break the cycle.
        """
        if self._sync_graph and not self._leaf_nodes:
            self._leaf_nodes = [
                obj
                for obj in self._sync_graph.leaf_nodes()
                if obj not in self._running_repos
            ]

            if not (self._leaf_nodes or self._running_repos):
                # If there is a circular master relationship,
                # choose a random node to break the cycle.
                self._leaf_nodes = [next(iter(self._sync_graph))]

    def _next_task(self):
        """
        Return a task for the next available leaf node.
        """
        if not self._sync_graph:
            raise StopIteration()
        # If self._sync_graph is non-empty, then self._leaf_nodes
        # is guaranteed to be non-empty, since otherwise
        # _can_add_job would have returned False and prevented
        # _next_task from being immediately called.
        node = self._leaf_nodes.pop()
        self._running_repos.add(node)
        self._update_leaf_nodes()

        return self._sync_manager.sync_async(
            emerge_config=self._emerge_config,
            repo=self._repo_map[node],
            master_hooks=self._master_hooks(node),
        )

    def _can_add_job(self):
        """
        Returns False if there are no leaf nodes available.
        """
        if not AsyncScheduler._can_add_job(self):
            return False
        return bool(self._leaf_nodes) and not self._terminated.is_set()

    def _keep_scheduling(self):
        """
        Schedule as long as the graph is non-empty, and we haven't
        been terminated.
        """
        return bool(self._sync_graph) and not self._terminated.is_set()



# Copyright 2018-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import subprocess

import portage
from portage.const import PORTAGE_PYM_PATH
from portage.tests import TestCase


class WakeupFdSigchldTestCase(TestCase):
    def testWakeupFdSigchld(self):
        """
        This is expected to trigger a bunch of messages like the following
        unless the fix for bug 655656 works as intended:

        Exception ignored when trying to write to the signal wakeup fd:
        BlockingIOError: [Errno 11] Resource temporarily unavailable
        """

        script = """
import os
import signal
import sys

import portage

# In order to avoid potential interference with API consumers, wakeup
# fd handling is enabled only when portage._interal_caller is True.
portage._internal_caller = True

from portage.util.futures import asyncio

loop = asyncio._wrap_loop()

# Cause the loop to register a child watcher.
proc = loop.run_until_complete(asyncio.create_subprocess_exec('sleep', '0', loop=loop))
loop.run_until_complete(proc.wait())

for i in range(8192):
	os.kill(portage.getpid(), signal.SIGCHLD)

# Verify that the child watcher still works correctly
# (this will hang if it doesn't).
proc = loop.run_until_complete(asyncio.create_subprocess_exec('sleep', '0', loop=loop))
loop.run_until_complete(proc.wait())
loop.close()
sys.stdout.write('success')
sys.exit(os.EX_OK)
"""

        pythonpath = os.environ.get("PYTHONPATH", "").strip().split(":")
        if not pythonpath or pythonpath[0] != PORTAGE_PYM_PATH:
            pythonpath = [PORTAGE_PYM_PATH] + pythonpath
        pythonpath = ":".join(filter(None, pythonpath))

        proc = subprocess.Popen(
            [portage._python_interpreter, "-c", script],
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=dict(os.environ, PYTHONPATH=pythonpath),
        )

        out, err = proc.communicate()
        try:
            self.assertEqual(out[:100], b"success")
        except Exception:
            portage.writemsg(
                "".join(
                    f"{line}\n"
                    for line in out.decode(errors="replace").splitlines()[:50]
                ),
                noiselevel=-1,
            )
            raise

        self.assertEqual(proc.wait(), os.EX_OK)



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Brian Harring (ferringb@gentoo.org)

import os as _os
from portage.cache import template
from portage import os

from portage.proxy.lazyimport import lazyimport

lazyimport(
    globals(),
    "portage.exception:PortageException",
    "portage.util:apply_permissions,ensure_dirs",
)
del lazyimport


class FsBased(template.database):
    """template wrapping fs needed options, and providing _ensure_access as a way to
    attempt to ensure files have the specified owners/perms"""

    def __init__(self, *args, **config):
        for x, y in (("gid", -1), ("perms", 0o644)):
            if x in config:
                # Since Python 3.4, chown requires int type (no proxies).
                setattr(self, "_" + x, int(config[x]))
                del config[x]
            else:
                setattr(self, "_" + x, y)
        super().__init__(*args, **config)

        if self.label.startswith(os.path.sep):
            # normpath.
            self.label = os.path.sep + os.path.normpath(self.label).lstrip(os.path.sep)

    def _ensure_access(self, path, mtime=-1):
        """returns true or false if it's able to ensure that path is properly chmod'd and chowned.
        if mtime is specified, attempts to ensure that's correct also"""
        try:
            apply_permissions(path, gid=self._gid, mode=self._perms)
            if mtime != -1:
                mtime = int(mtime)
                os.utime(path, (mtime, mtime))
        except (PortageException, OSError):
            return False
        return True

    def _ensure_dirs(self, path=None):
        """with path!=None, ensure beyond self.location.  otherwise, ensure self.location"""
        if path:
            path = os.path.dirname(path)
            base = self.location
        else:
            path = self.location
            base = "/"

        for d in path.lstrip(os.path.sep).rstrip(os.path.sep).split(os.path.sep):
            base = os.path.join(base, d)
            if ensure_dirs(base):
                # We only call apply_permissions if ensure_dirs created
                # a new directory, so as not to interfere with
                # permissions of existing directories.
                mode = self._perms
                if mode == -1:
                    mode = 0
                mode |= 0o755
                apply_permissions(base, mode=mode, gid=self._gid)

    def _prune_empty_dirs(self):
        all_dirs = []
        for parent, dirs, files in os.walk(self.location):
            for x in dirs:
                all_dirs.append(_os.path.join(parent, x))
        while all_dirs:
            try:
                _os.rmdir(all_dirs.pop())
            except OSError:
                pass


def gen_label(base, label):
    """if supplied label is a path, generate a unique label based upon label, and supplied base path"""
    if label.find(os.path.sep) == -1:
        return label
    label = label.strip('"').strip("'")
    label = os.path.join(*(label.rstrip(os.path.sep).split(os.path.sep)))
    tail = os.path.split(label)[1]
    return f"{tail}-{abs(label.__hash__()):X}"



#!/usr/bin/env python
# Copyright 2006-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import grp
import os
import os.path as osp
import pwd
import signal
import tempfile
import shutil
import sys

import pytest

import portage
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.const import PORTAGE_BIN_PATH


def debug_signal(signum, frame):
    import pdb

    pdb.set_trace()


signal.signal(signal.SIGUSR1, debug_signal)


@pytest.fixture(autouse=True, scope="session")
def prepare_environment():
    # Pretend that the current user's uid/gid are the 'portage' uid/gid,
    # so things go smoothly regardless of the current user and global
    # user/group configuration.
    os.environ["PORTAGE_USERNAME"] = pwd.getpwuid(os.getuid()).pw_name
    os.environ["PORTAGE_GRPNAME"] = grp.getgrgid(os.getgid()).gr_name

    # Insert our parent dir so we can do shiny import "tests"
    # This line courtesy of Marienz and Pkgcore ;)
    sys.path.insert(0, osp.dirname(osp.dirname(osp.dirname(osp.realpath(__file__)))))

    portage._internal_caller = True

    # Ensure that we don't instantiate portage.settings, so that tests should
    # work the same regardless of global configuration file state/existence.
    portage._disable_legacy_globals()

    if portage.util.no_color(os.environ):
        portage.output.nocolor()

    # import portage.tests as tests

    path = os.environ.get("PATH", "").split(":")
    path = [x for x in path if x]

    insert_bin_path = True
    try:
        insert_bin_path = not path or not os.path.samefile(path[0], PORTAGE_BIN_PATH)
    except OSError:
        pass

    if insert_bin_path:
        path.insert(0, PORTAGE_BIN_PATH)
        os.environ["PATH"] = ":".join(path)

    try:
        # Copy GPG test keys to temporary directory
        gpg_path = tempfile.mkdtemp(prefix="gpg_")

        shutil.copytree(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), ".gnupg"),
            gpg_path,
            dirs_exist_ok=True,
        )

        os.chmod(gpg_path, 0o700)
        os.environ["PORTAGE_GNUPGHOME"] = gpg_path

        yield

    finally:
        global_event_loop().close()
        shutil.rmtree(gpg_path, ignore_errors=True)


# if __name__ == "__main__":
#     try:
#         sys.exit(tests.main())
#     finally:
#         global_event_loop().close()
#         shutil.rmtree(gpg_path, ignore_errors=True)



# Copyright 2017-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class VirtualMinimizeChildrenTestCase(TestCase):
    def testVirtualMinimizeChildren(self):
        ebuilds = {
            "app-misc/bar-1": {"EAPI": "6", "RDEPEND": "virtual/foo"},
            "virtual/foo-1": {
                "EAPI": "6",
                "RDEPEND": "|| ( app-misc/A app-misc/B ) || ( app-misc/B app-misc/C )",
            },
            "app-misc/A-1": {
                "EAPI": "6",
            },
            "app-misc/B-1": {
                "EAPI": "6",
            },
            "app-misc/C-1": {
                "EAPI": "6",
            },
        }

        test_cases = (
            # Test bug 632026, where we want to minimize the number of
            # packages chosen to satisfy overlapping || deps like
            # "|| ( foo bar ) || ( bar baz )".
            ResolverPlaygroundTestCase(
                ["app-misc/bar"],
                success=True,
                mergelist=[
                    "app-misc/B-1",
                    "virtual/foo-1",
                    "app-misc/bar-1",
                ],
            ),
        )

        playground = ResolverPlayground(debug=False, ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

        # If app-misc/A and app-misc/C are installed then
        # that choice should be preferred over app-misc/B.
        installed = {
            "app-misc/A-1": {
                "EAPI": "6",
            },
            "app-misc/C-1": {
                "EAPI": "6",
            },
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/bar"],
                success=True,
                mergelist=[
                    "virtual/foo-1",
                    "app-misc/bar-1",
                ],
            ),
        )

        playground = ResolverPlayground(
            debug=False, ebuilds=ebuilds, installed=installed
        )

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def testOverlapSlotConflict(self):
        ebuilds = {
            "app-misc/bar-1": {"EAPI": "6", "RDEPEND": "virtual/foo"},
            "virtual/foo-1": {
                "EAPI": "6",
                "RDEPEND": "|| ( app-misc/A >=app-misc/B-2 ) || ( <app-misc/B-2 app-misc/C )",
            },
            "app-misc/A-1": {
                "EAPI": "6",
            },
            "app-misc/B-2": {
                "EAPI": "6",
            },
            "app-misc/B-1": {
                "EAPI": "6",
            },
            "app-misc/C-1": {
                "EAPI": "6",
            },
        }

        test_cases = (
            # Here the ( >=app-misc/B-2 <app-misc/B-2 ) choice is not satisfiable.
            ResolverPlaygroundTestCase(
                ["app-misc/bar"],
                success=True,
                ambiguous_merge_order=True,
                mergelist=[
                    (
                        "app-misc/C-1",
                        "app-misc/A-1",
                    ),
                    "virtual/foo-1",
                    "app-misc/bar-1",
                ],
            ),
        )

        playground = ResolverPlayground(debug=False, ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def testVirtualPackageManager(self):
        ebuilds = {
            "app-admin/perl-cleaner-2.25": {
                "RDEPEND": """
					|| (
						( sys-apps/portage app-portage/portage-utils )
						sys-apps/pkgcore
						sys-apps/paludis
					)"""
            },
            "app-portage/portage-utils-0.64": {},
            "sys-apps/paludis-2.6.0": {},
            "sys-apps/portage-2.3.19-r1": {},
            "virtual/package-manager-0": {
                "RDEPEND": """
					|| (
						sys-apps/portage
						sys-apps/paludis
						sys-apps/pkgcore
					)"""
            },
        }

        test_cases = (
            # Test bug 645002, where paludis was selected to satisfy a
            # perl-cleaner dependency because that choice contained fewer
            # packages than the ( portage portage-utils ) choice which
            # should have been preferred according to the order of
            # choices specified in the ebuild.
            ResolverPlaygroundTestCase(
                [
                    "app-admin/perl-cleaner",
                    "virtual/package-manager",
                ],
                all_permutations=True,
                success=True,
                ambiguous_merge_order=True,
                mergelist=(
                    (
                        "sys-apps/portage-2.3.19-r1",
                        "app-portage/portage-utils-0.64",
                        "app-admin/perl-cleaner-2.25",
                        "virtual/package-manager-0",
                    ),
                ),
            ),
            # Test paludis preference. In this case, if paludis is not
            # included in the argument atoms then the result varies
            # depending on whether the app-admin/perl-cleaner or
            # virtual/package-manager dependencies are evaluated first!
            # Therefore, include paludis in the argument atoms.
            ResolverPlaygroundTestCase(
                [
                    "app-admin/perl-cleaner",
                    "virtual/package-manager",
                    "sys-apps/paludis",
                ],
                all_permutations=True,
                success=True,
                ambiguous_merge_order=True,
                mergelist=(
                    "sys-apps/paludis-2.6.0",
                    (
                        "app-admin/perl-cleaner-2.25",
                        "virtual/package-manager-0",
                    ),
                ),
            ),
        )

        playground = ResolverPlayground(debug=False, ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def testVirtualDevManager(self):
        ebuilds = {
            "sys-fs/eudev-3.1.5": {},
            "sys-fs/static-dev-0.1": {},
            "sys-fs/udev-233": {},
            "virtual/dev-manager-0": {
                "RDEPEND": """
					|| (
						virtual/udev
						sys-fs/static-dev
					)"""
            },
            "virtual/udev-0": {
                "RDEPEND": """
					|| (
						>=sys-fs/eudev-2.1.1
						>=sys-fs/udev-217
					)"""
            },
        }

        test_cases = (
            # Test bug 645190, where static-dev was pulled in instead
            # of eudev.
            ResolverPlaygroundTestCase(
                [
                    "virtual/dev-manager",
                ],
                success=True,
                mergelist=(
                    "sys-fs/eudev-3.1.5",
                    "virtual/udev-0",
                    "virtual/dev-manager-0",
                ),
            ),
            # Test static-dev preference.
            ResolverPlaygroundTestCase(
                [
                    "sys-fs/static-dev",
                    "virtual/dev-manager",
                ],
                all_permutations=True,
                success=True,
                mergelist=(
                    "sys-fs/static-dev-0.1",
                    "virtual/dev-manager-0",
                ),
            ),
        )

        playground = ResolverPlayground(debug=False, ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()

    def testVirtualWine(self):
        ebuilds = {
            "virtual/wine-0-r6": {
                "RDEPEND": "|| ( app-emulation/wine-staging app-emulation/wine-any ) "
                "|| ( app-emulation/wine-vanilla app-emulation/wine-staging app-emulation/wine-any )"
            },
            "app-emulation/wine-staging-4": {},
            "app-emulation/wine-any-4": {},
            "app-emulation/wine-vanilla-4": {},
        }

        test_cases = (
            # Test bug 701996, where separate disjunctions where not
            # converted to DNF, causing both wine-vanilla and
            # wine-staging to be pulled in.
            ResolverPlaygroundTestCase(
                [
                    "virtual/wine",
                ],
                success=True,
                mergelist=(
                    "app-emulation/wine-staging-4",
                    "virtual/wine-0-r6",
                ),
            ),
        )

        playground = ResolverPlayground(debug=False, ebuilds=ebuilds)

        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage.tests import TestCase


class PreloadPortageSubmodulesTestCase(TestCase):
    def testPreloadPortageSubmodules(self):
        """
        Verify that _preload_portage_submodules() doesn't leave any
        remaining proxies that refer to the portage.* namespace.
        """
        portage.proxy.lazyimport._preload_portage_submodules()
        for name in portage.proxy.lazyimport._module_proxies:
            self.assertEqual(name.startswith("portage."), False)



# Copyright 2018-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import difflib
import os
import subprocess

import portage
from portage.const import PORTAGE_BIN_PATH
from portage.tests import TestCase


class TestFilterBashEnv(TestCase):
    def testTestFilterBashEnv(self):
        test_cases = (
            (
                "RDEPEND BASH.* _EPATCH_ECLASS",
                rb"""declare -ir BASHPID="28997"
declare -rx A="portage-2.3.24.tar.bz2"
declare -- DESKTOP_DATABASE_DIR="/usr/share/applications"
declare PDEPEND="
        !build? (
                >=net-misc/rsync-2.6.4
                userland_GNU? ( >=sys-apps/coreutils-6.4 )
        ) "
declare RDEPEND="
        >=app-arch/tar-1.27
        dev-lang/python-exec:2"
declare -x PF="portage-2.3.24"
declare -a PYTHON_COMPAT=([0]="pypy" [1]="python3_4" [2]="python3_5" [3]="python3_6" [4]="python2_7")
declare -- _EPATCH_ECLASS="1"
declare -- _EUTILS_ECLASS="1"
declare -- f
get_libdir ()
{
    local libdir_var="LIBDIR_${ABI}";
    local libdir="lib";
    [[ -n ${ABI} && -n ${!libdir_var} ]] && libdir=${!libdir_var};
    echo "${libdir}"
}
make_wrapper ()
{
    cat  <<-EOF
export ${var}="\${${var}}:${EPREFIX}${libdir}"
EOF
}
use_if_iuse ()
{
    in_iuse $1 || return 1;
    use $1
}
""",
                rb"""declare -x A="portage-2.3.24.tar.bz2"
declare -- DESKTOP_DATABASE_DIR="/usr/share/applications"
declare PDEPEND="
        !build? (
                >=net-misc/rsync-2.6.4
                userland_GNU? ( >=sys-apps/coreutils-6.4 )
        ) "
declare -x PF="portage-2.3.24"
declare -a PYTHON_COMPAT=([0]="pypy" [1]="python3_4" [2]="python3_5" [3]="python3_6" [4]="python2_7")
declare -- _EUTILS_ECLASS="1"
declare -- f
get_libdir ()
{
    local libdir_var="LIBDIR_${ABI}";
    local libdir="lib";
    [[ -n ${ABI} && -n ${!libdir_var} ]] && libdir=${!libdir_var};
    echo "${libdir}"
}
make_wrapper ()
{
    cat  <<-EOF
export ${var}="\${${var}}:${EPREFIX}${libdir}"
EOF
}
use_if_iuse ()
{
    in_iuse $1 || return 1;
    use $1
}
""",
            ),
        )

        for filter_vars, env_in, env_out in test_cases:
            proc = None
            try:
                proc = subprocess.Popen(
                    [
                        portage._python_interpreter,
                        os.path.join(PORTAGE_BIN_PATH, "filter-bash-environment.py"),
                        filter_vars,
                    ],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                )
                proc.stdin.write(env_in)
                proc.stdin.close()
                result = proc.stdout.read()
            finally:
                if proc is not None:
                    proc.stdin.close()
                    proc.wait()
                    proc.stdout.close()

            diff = list(
                difflib.unified_diff(
                    env_out.decode("utf_8").splitlines(),
                    result.decode("utf_8").splitlines(),
                )
            )

            self.assertEqual(diff, [])



# Copyright 2019-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import asyncio
import functools
import os
import shutil
import socket
import struct
import subprocess
import tempfile
import time

import portage
from portage.tests import TestCase, get_pythonpath
from portage.util import socks5
from portage.util.futures.executor.fork import ForkExecutor
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.const import PORTAGE_BIN_PATH

from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.request import urlopen


class _Handler(BaseHTTPRequestHandler):
    def __init__(self, content, *args, **kwargs):
        self.content = content
        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def do_GET(self):
        doc = self.send_head()
        if doc is not None:
            self.wfile.write(doc)

    def do_HEAD(self):
        self.send_head()

    def send_head(self):
        doc = self.content.get(self.path)
        if doc is None:
            self.send_error(404, "File not found")
            return None

        self.send_response(200)
        self.send_header("Content-type", "text/plain")
        self.send_header("Content-Length", len(doc))
        self.send_header("Last-Modified", self.date_time_string(time.time()))
        self.end_headers()
        return doc

    def log_message(self, fmt, *args):
        pass


class AsyncHTTPServer:
    def __init__(self, host, content, loop):
        self._host = host
        self._content = content
        self._loop = loop
        self.server_port = None
        self._httpd = None

    def pause(self):
        """Pause responses (useful for testing timeouts)."""
        self._loop.remove_reader(self._httpd.socket.fileno())

    def resume(self):
        """Resume responses following a previous call to pause."""
        self._loop.add_reader(
            self._httpd.socket.fileno(), self._httpd._handle_request_noblock
        )

    def __enter__(self):
        httpd = self._httpd = HTTPServer(
            (self._host, 0), functools.partial(_Handler, self._content)
        )
        self.server_port = httpd.server_port
        self.resume()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self._httpd is not None:
            self.pause()
            self._httpd.socket.close()
            self._httpd = None


class AsyncHTTPServerTestCase(TestCase):
    @staticmethod
    def _fetch_directly(host, port, path):
        # NOTE: python2.7 does not have context manager support here
        try:
            f = urlopen(
                "http://{host}:{port}{path}".format(  # nosec
                    host=host, port=port, path=path
                )
            )
            return f.read()
        finally:
            if f is not None:
                f.close()

    async def _test_http_server(self):
        asyncio.run(self._test_http_server())

    async def _test_http_server(self):
        host = "127.0.0.1"
        content = b"Hello World!\n"
        path = "/index.html"

        loop = asyncio.get_running_loop()
        for i in range(2):
            with AsyncHTTPServer(host, {path: content}, loop) as server:
                for j in range(2):
                    result = await loop.run_in_executor(
                        None, self._fetch_directly, host, server.server_port, path
                    )
                    self.assertEqual(result, content)


class _socket_file_wrapper(portage.proxy.objectproxy.ObjectProxy):
    """
    A file-like object that wraps a socket and closes the socket when
    closed. Since python2.7 does not support socket.detach(), this is a
    convenient way to have a file attached to a socket that closes
    automatically (without resource warnings about unclosed sockets).
    """

    __slots__ = ("_file", "_socket")

    def __init__(self, socket, f):
        object.__setattr__(self, "_socket", socket)
        object.__setattr__(self, "_file", f)

    def _get_target(self):
        return object.__getattribute__(self, "_file")

    def __getattribute__(self, attr):
        if attr == "close":
            return object.__getattribute__(self, "close")
        return super().__getattribute__(attr)

    def __enter__(self):
        return self

    def close(self):
        object.__getattribute__(self, "_file").close()
        object.__getattribute__(self, "_socket").close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def socks5_http_get_ipv4(proxy, host, port, path):
    """
    Open http GET request via socks5 proxy listening on a unix socket,
    and return a file to read the response body from.
    """
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    f = _socket_file_wrapper(s, s.makefile("rb", 1024))
    try:
        s.connect(proxy)
        s.send(struct.pack("!BBB", 0x05, 0x01, 0x00))
        vers, method = struct.unpack("!BB", s.recv(2))
        s.send(struct.pack("!BBBB", 0x05, 0x01, 0x00, 0x01))
        s.send(socket.inet_pton(socket.AF_INET, host))
        s.send(struct.pack("!H", port))
        reply = struct.unpack("!BBB", s.recv(3))
        if reply != (0x05, 0x00, 0x00):
            raise AssertionError(repr(reply))
        struct.unpack("!B4sH", s.recv(7))  # contains proxied address info
        s.send(
            "GET {} HTTP/1.1\r\nHost: {}:{}\r\nAccept: */*\r\nConnection: close\r\n\r\n".format(
                path, host, port
            ).encode()
        )
        headers = []
        while True:
            headers.append(f.readline())
            if headers[-1] == b"\r\n":
                return f
    except Exception:
        f.close()
        raise


class Socks5ServerTestCase(TestCase):
    @staticmethod
    def _fetch_via_proxy(proxy, host, port, path):
        with socks5_http_get_ipv4(proxy, host, port, path) as f:
            return f.read()

    def test_socks5_proxy(self):
        asyncio.run(self._test_socks5_proxy())

    async def _test_socks5_proxy(self):
        loop = global_event_loop()

        host = "127.0.0.1"
        content = b"Hello World!"
        path = "/index.html"
        proxy = None
        tempdir = tempfile.mkdtemp()
        previous_exithandlers = loop._coroutine_exithandlers

        try:
            loop._coroutine_exithandlers = []
            with AsyncHTTPServer(host, {path: content}, loop) as server:
                settings = {
                    "PORTAGE_TMPDIR": tempdir,
                    "PORTAGE_BIN_PATH": PORTAGE_BIN_PATH,
                }

                proxy = socks5.get_socks5_proxy(settings)
                await socks5.proxy.ready()

                result = await loop.run_in_executor(
                    None,
                    self._fetch_via_proxy,
                    proxy,
                    host,
                    server.server_port,
                    path,
                )

                self.assertEqual(result, content)
        finally:
            try:
                # Also run_coroutine_exitfuncs to test atexit hook cleanup.
                self.assertNotEqual(loop._coroutine_exithandlers, [])
                await portage.process.run_coroutine_exitfuncs()
                self.assertEqual(loop._coroutine_exithandlers, [])
            finally:
                loop._coroutine_exithandlers = previous_exithandlers
                shutil.rmtree(tempdir)


class Socks5ServerLoopCloseTestCase(TestCase):
    """
    For bug 925240, test that the socks5 proxy is automatically
    terminated when the main event loop is closed, using a subprocess
    for isolation.
    """

    def testSocks5ServerLoopClose(self):
        asyncio.run(self._testSocks5ServerLoopClose())

    async def _testSocks5ServerLoopClose(self):
        loop = asyncio.get_running_loop()
        self.assertEqual(
            await loop.run_in_executor(
                ForkExecutor(loop=loop), self._testSocks5ServerLoopCloseSubprocess
            ),
            True,
        )

    @staticmethod
    def _testSocks5ServerLoopCloseSubprocess():
        loop = global_event_loop()
        tempdir = tempfile.mkdtemp()
        try:
            settings = {
                "PORTAGE_TMPDIR": tempdir,
                "PORTAGE_BIN_PATH": PORTAGE_BIN_PATH,
            }

            socks5.get_socks5_proxy(settings)
            loop.run_until_complete(socks5.proxy.ready())
        finally:
            loop.close()
            shutil.rmtree(tempdir)

        return not socks5.proxy.is_running()


class Socks5ServerAtExitTestCase(TestCase):
    """
    For bug 937384, test that the socks5 proxy is automatically
    terminated by portage.process.run_exitfuncs(), using a subprocess
    for isolation.

    Note that if the subprocess is created via fork then it will be
    vulnerable to python issue 83856 which is only fixed in python3.13,
    so this test uses python -c to ensure that atexit hooks will work.
    """

    _threaded = False

    def testSocks5ServerAtExit(self):
        tempdir = tempfile.mkdtemp()
        try:
            env = os.environ.copy()
            env["PYTHONPATH"] = get_pythonpath()
            output = subprocess.check_output(
                [
                    portage._python_interpreter,
                    "-c",
                    """
import sys
import threading

from portage.const import PORTAGE_BIN_PATH
from portage.util import socks5
from portage.util._eventloop.global_event_loop import global_event_loop

tempdir = sys.argv[0]
threaded = bool(sys.argv[1])

settings = {
    "PORTAGE_TMPDIR": tempdir,
    "PORTAGE_BIN_PATH": PORTAGE_BIN_PATH,
}

def main():
    loop = global_event_loop()
    socks5.get_socks5_proxy(settings)
    loop.run_until_complete(socks5.proxy.ready())
    print(socks5.proxy._proc.pid, flush=True)

if __name__ == "__main__":
    if threaded:
        t = threading.Thread(target=main)
        t.start()
        t.join()
    else:
        main()
""",
                    tempdir,
                    str(self._threaded),
                ],
                env=env,
            )

            pid = int(output.strip())

            with self.assertRaises(ProcessLookupError):
                os.kill(pid, 0)
        finally:
            shutil.rmtree(tempdir)


class Socks5ServerAtExitThreadedTestCase(Socks5ServerAtExitTestCase):
    _threaded = True
